import math
import numpy

from PyQt5 import QtCore, QtGui

from API import BuilderPanelPlugin
from QuantumATK import Angstrom, BulkConfiguration

from NL.GUI.Presenter.Presenter import Presenter
from NL.CommonConcepts.Configurations.ConfigurationCopy import configurationCopy

from .UI.BucklerWidget import BucklerWidget

def buckle(x, width, length, amplitude, z_start, z_end, periods_y, periods_z):
    """
    Function for converting a nanosheet coordinate into a wobbled graphene nano-sheet
    @param x         : Coordinates of an atom
    @param width     : wave length of the buckling in the y-direction
    @param length    : wave length of the buckling in the z-direction
    @param amplitude : Buckling in units of Angstrom
    @param z_start   : z value for starting the buckling
    @param z_end     : z value for ending the buckling
    @param periods_y : number of periods in y
    @param periods_z : number of periods in z
    """
    # do not buckle for z > z_end
    z = x[2]
    z = min(z,z_end)
    # do not buckle for z < z_start
    z = z - z_start
    z = max (z,0.0)
    # find the buckling phase of the current atom in the y and z directions
    if periods_y==0:
        phase_y = math.pi/2
    else:
        phase_y = 2.*math.pi*(x[1]-width/2.)/width * periods_y
    if periods_z==0:
        phase_z = math.pi/2
    else:
        phase_z = 2.*math.pi*(z-length/2.)/length * periods_z

    # return atom position of the wrapped atom
    return x + numpy.array([amplitude*math.sin(phase_y)*math.sin(phase_z),0.0,0.0])

class BucklerPlugin(BuilderPanelPlugin):
    """
    Plugin for buckling
    """
    def __init__(self, parent=None):
        """
        Constructor.

        @parent : The widget that holds the plugin.
        """
        self._widget = BucklerWidget()

        BuilderPanelPlugin.__init__(self, parent)

        # Placeholder for the preview
        self._preview = None

        self._widget.apply_button.clicked.connect(self.applyClicked)
        self._widget.preview_button.clicked.connect(self.preview)
        self._widget.amplitude.valueChanged[float].connect(self.preview)
        self._widget.nonbucklingzone.valueChanged[float].connect(self.preview)
        self._widget.periods_y.valueChanged[int].connect(self.preview)
        self._widget.periods_z.valueChanged[int].connect(self.preview)

    def title(self):
        """
        @return The title of the panel.
        """
        return self.tr("Buckler")

    def widget(self):
        """
        @return The widget presented to the user.
        """
        return self._widget

    @staticmethod
    def group():
        return 'Coordinate Tools'

    def doBuckling(self):
        """
        Returns the buckled coordinates of the active configuration
        """
        # Get active presenter and configuration
        presenter = self.stash().activePresenter()
        configuration = presenter.configuration()

        # Parameters from the widget
        amplitude = self._widget.amplitude.value()
        nonbucklingzone = self._widget.nonbucklingzone.value()
        periods_y = self._widget.periods_y.value()
        periods_z = self._widget.periods_z.value()

        # Lengths of the unit vectors
        lattice = configuration.bravaisLattice()
        y_length = numpy.linalg.norm(lattice.primitiveVectors()[1].inUnitsOf(Angstrom))
        z_length = numpy.linalg.norm(lattice.primitiveVectors()[2].inUnitsOf(Angstrom))

        # Define a function of one variable, f(c), for displacing the atoms
        f = lambda c : buckle(c, y_length, z_length-nonbucklingzone*2.,
                              amplitude, nonbucklingzone, z_length-nonbucklingzone,
                              periods_y, periods_z)
        # Apply the function to find new displaced coordinates
        buckled_coordinates = numpy.apply_along_axis(f, 1, configuration.cartesianCoordinates().inUnitsOf(Angstrom))

        return buckled_coordinates

    def applyClicked(self):
        """
        Called when the 'Apply' button is pressed.
        """
        # Get active presenter and configuration
        presenter = self.stash().activePresenter()
        configuration = presenter.configuration()

        # Get buckled coordinates
        new_coordinates = self.doBuckling()

        # Set undo, then give new config to presenter
        presenter.setUndoMark('Buckle')
        presenter.changeAtoms(positions=new_coordinates*Angstrom)

        # Remove preview
        self.stash().setPreview(None, self)

    def preview(self):
        """
        Called when the 'Preview' button is pressed or any value is changed.
        """
        # Get buckled coordinates
        new_coordinates = self.doBuckling()

        self._preview.changeAtoms(positions=new_coordinates*Angstrom)
        self.stash().setPreview(self._preview, self)

    def activeConfigurationModified(self):
        """
           Called when changes occur in the stash of configurations.
        """
        # Set enabled state based on configuration
        configuration = self.stash().activeConfiguration()
        if isinstance(configuration, BulkConfiguration) and len(configuration) > 0:
            self.setAvailable(True)
            # Make a local copy of the configuration to use for the preview
            if self._preview == None:
                self._preview = Presenter(configurationCopy(configuration))
            else:
                self._preview.setConfiguration(configurationCopy(configuration))
        else:
            self.setAvailable(False)
